<?
# dvbdcontrol
#
# Web-interface to control the dvbd tools. 
# Use only in trusted networks or with additional authentication!
#
# Copyright (c) 2005 Tim Jansen
# See LICENSE for licensing (MIT license).
# 


require 'config.php';

# Paths for the tools
$dvbsched = $dvbBinPath . 'dvbsched';
$dvblist = $dvbBinPath . 'dvblist';
$dvbcat = $dvbBinPath . 'dvbcat';

# Recording list
$recordingListPath = $recordDir . '.webdvbd-recording-list';

# Priority Constants
$PRIO_LOW = 5;
$PRIO_NORMAL = 10;
$PRIO_HIGH = 20;

# Internal function to parse whitespace-escaped strings from dvbd
function unescapeWS($s) {
	$result = "";
	$len = strlen($s);
	$pos = 0;
	while (true) {
		$npos = strpos($s, "\\", $pos);
		if ($npos === false)
			return $result . substr($s, $pos);
		$result .= substr($s, $pos, $npos - $pos);
		if ($npos < $len - 1) {
			if ($s{$npos+1} == "\\") {
				$result .= "\\";
				$pos = $npos + 2;
			}
			else if ($s{$npos+1} == "_") {
				$result .= ' ';
				$pos = $npos + 2;
			}
			else {
				$result .= "\\";
				$pos = $npos + 1;
			}
		}
		else
                        return $result . "\\";
	}
}

# Internal function to create whitespace-escaped strings like dvbd's
function escapeWS($s) {
	$result = "";
	$len = strlen($s);
	$pos = 0;
	while (true) {
		$npos = strpos($s, " ", $pos);
		$npos2 = strpos($s, "\\", $pos);
		if ($npos === false || ($npos2 === true && $npos2 < $npos)) {
			if ($npos2 === false)
				return $result . substr($s, $pos);
			else {
				$result .= substr($s, $pos, $npos2 - $pos);
				$result .= "\\\\";
				$pos = $npos2 + 1;
			}
		}
		else {
			$result .= substr($s, $pos, $npos - $pos);
			$result .= "\_";
			$pos = $npos + 1;
		}
	}
}

# Internal function to add time in the dvbd format
function addTime($timeStr, $durationMin) {
	$timeMin = intval(substr($timeStr, 0, 2))*60 + 
		   intval(substr($timeStr, 3, 2)) +
		   $durationMin;
	while ($timeMin < 0)
		$timeMin += 60*24;
	while ($timeMin >= 60*24)
		$timeMin -= 60*24;
	return sprintf("%02d:%02d", $timeMin/60, $timeMin%60);
}

# Internal function to get the time difference in minutes 
# between two times in in the dvbd format. If timeEndStr is
# lower, it assumed that it is after midnight and thus wraps.
function timeDiff($timeBeginStr, $timeEndStr) {
	$timeBegin = intval(substr($timeBeginStr, 0, 2))*60 +
		     intval(substr($timeBeginStr, 3, 2));
	$timeEnd   = intval(substr($timeEndStr, 0, 2))*60 +
		     intval(substr($timeEndStr, 3, 2));

	if ($timeBegin > $timeEnd)
		$timeEnd += 60*24;

	return $timeEnd - $timeBegin;
}

# Internal function to parse a job
function readJob($line) {
	global $recordDir;
        $cs = explode(" ", $line, 12);
	$file = unescapeWS($cs[10]);
	if (strpos($file, $recordDir) === 0)
		$file = substr($file,  strlen($recordDir));
	$duration = intval($cs[8])/60;
	$time = substr($cs[4], 0, 5);
	return array(   'number' => intval($cs[1]),
                        'user' => intval($cs[2]),
                        'date' => $cs[3],
                        'time' => $time,
			'endTime' => addTime($time, $duration),
                        'dvbType' => $cs[5],
                        'channel' => unescapeWS($cs[6]),
                        'convertToPs' => $cs[7]=='true',
                        'duration' => $duration,
                        'priority' => intval($cs[9]),
                        'file' => $file,
                        'extra' => unescapeWS(substr($cs[11], 1, strlen($cs[11])-2))
                );
}

# Internal function to parse a connection, returns array map containing
#   'priority' => Priority number
#   'dvbType' => Connection type, 'dvb-t', 'dvb-s' or 'dvb-c'
#   'channel' => Channel name (see getChannels())
#   'job'     => If it is a recording, a job map like from getSchedule(),
#                otherwise null
function readConnection($line) {
	global $recordDir;
        $cs = explode(" ", $line, 5);
	$c = array(   'priority' => intval($cs[1]),
                        'dvbType' => $cs[2],
                        'channel' => unescapeWS($cs[3])
                );
	if ($cs[4] != null && strlen($cs[4]) > 5)
		$c['job'] = readJob(' '.$cs[4]);
	else
		$c['job'] = null;
	return $c;
}

# Returns a list of all scheduled recordings (or null for error). 
# List is a numbered array, ordered by start time, beginning with the next
# job. Each entry is a array map that contains the details of the job:
#   'number' => Job number (needed for removing)
#   'user' => Unix (/etc/passwd) uid of the user
#   'date' => Date of the job, yyyy-mm-dd, e.g. '2005-02-21'
#   'time' => Time of the job, hh:mm, e.g. '09:57'
#   'endTime' => End of the job, hh:mm, e.g. '11:11'
#   'dvbType' => Recording type, 'dvb-t', 'dvb-s' or 'dvb-c'
#   'channel' => Channel name (see getChannels())
#   'convertToPs' => 'true' to record program stream, 'false' for TS
#   'duration' => The duration of the recording in minutes
#   'priority' => Priority of the job (number), higher preferred, default is 10
#   'file' => Absolute path of the file to write
#   'extra' => String of arbitrary extra infos as specified by the user, 
function getSchedule() {
	global $dvblist;
	$schedule = shell_exec("$dvblist -h");
	if (!$schedule)
		return null;
	$lines = explode("\n", $schedule);
	array_shift($lines);               # remove first line ('schedule is')
	array_pop($lines);                 # remove last line (after last '\n')

	$recordings = array();
	foreach ($lines as $number => $line)
		$recordings[] = readJob($line);

	return $recordings;
}

# Retrieves the job with the given number from the
# schedule list.
# Returns the job array (see getSchedule()), or null
# if not found.
function getJob($number) {
	$schedule = getSchedule();
	foreach ($schedule as $key => $job) {
		if ($job['number'] == $number)
			return $job;
	}
	return null;
}

# Returns a list of all channel names (at priority 1000000).
# Return value is a numbered array, each channel is a string,
# ordered alphabetically.
function getChannels() {
	return getChannelsByPriority(1000000);
}


# Returns a list of the channel names that can be tunes to with 
# the given priority.
# Return value is a numbered array, each channel is a string,
# ordered alphabetically.
function getChannelsByPriority($priority)
{
	global $dvblist, $dvbType;

	$p = escapeshellarg($priority);
        $list = shell_exec("$dvblist -p $p $dvbType");
        $channels = explode("\n", $list);
        array_shift($channels);
        array_pop($channels);
        return $channels;
}

# Returns a list of current connections. Empty list if there is no activity.
# Returns a list of connections, null for error. Each connection is an array map:
#   'priority' => Priority number
#   'dvbType'  => Connection type, 'dvb-t', 'dvb-s' or 'dvb-c'
#   'channel'  => Channel name (see getChannels())
#   'job'      => If it is a recording, a job map like from getSchedule(),
#                 otherwise 0
function getConnections() 
{
	global $dvblist;
	$connections = shell_exec("$dvblist -c 2>&1");
	if (!$connections)
		return null;
	$lines = explode("\n", $connections);
	array_shift($lines);               # remove first line ('Current connections'...)
	array_pop($lines);                 # remove last line (after last '\n')

	$connections = array();
	$rcbegin = "RecordConnection";
	$ccbegin = "ControlledConnection";
	foreach ($lines as $number => $line) {
		if (strpos($line, $rcbegin) === 0)
			$connections[] = readConnection(substr($line, strlen($rcbegin)));
		if (strpos($line, $ccbegin) === 0 &&
		    strpos($line, "$ccbegin UNTUNED") === false)
			$connections[] = readConnection(substr($line, strlen($ccbegin)));
	}
	return $connections;
}


# Removes the job with the given number. Returns null on success or an
# error message on failure.
function removeJob($number) {
	global $dvbsched;
	$number = intval($number);
	$output = shell_exec("$dvbsched -r $number 2>&1");
	if ($output === false)
		return "Can not execute $dvbsched";
	if ($output == null) {
		updateRecordingList();
		return null;
	}
	$msgbegin = "Failed to remove job: ";
	if (strpos($output, $msgbegin) !== 0)
		return "Got unparsable output from $dvbsched:\n $output";
	return substr($output, strlen($msgbegin));
}

# Adds the job to the schedule. 
# Argument is a job map, like the one returned in the getSchedule() list.
# 'number', 'user', 'dvbType' will be ignored. If 'file' does not start with
# a slash ('/'), $recordDir from config.php will be prepended. If 'endTime' is
# set, it will be taken, otherwise 'duration'.
# Returns an array map with two values:
#   'result': null on success, or an error message on failure.
#   'number': the new number of the job after editing. Only set on success
function addJob($job) {
	global $dvbsched, $dvbType, $recordDir;

	$file = $job['file'];
	if (strpos($file, "/") !== 0)
		$file = $recordDir . $file;
	$file = escapeshellarg($file);

	$priority = intval($job['priority']);
	$ps = $job['convertToPs'] ? '--ps' : '';
	$extra = strlen($job['extra'])>0 ? '-e '.escapeshellarg('['.$job['extra']/']') : "";
	$channel = escapeshellarg($job['channel']);
	$datetime = escapeshellarg($job['time'] . ' ' . $job['date']);
	if ($job['endTime'] != null)
		$duration = timeDiff($job['time'], $job['endTime']);
	else
		$duration = intval($job['duration']);
	$cmd = "$dvbsched -n --priority $priority $ps $extra $dvbType $channel $datetime $duration $file 2>&1";
	$output = shell_exec($cmd);
	if (!$output)
		return array('result' => "Can not execute $cmd");
        $lines = explode("\n", $output);
        array_shift($lines);
        array_pop($lines);
	$line = $lines[0];

	$msgbegin = "Failed to schedule job: ";
	if (strpos($line, $msgbegin) === 0) {
		if (strpos($line, 'SCHEDULE PATH ACCESS ERROR') >= 0)
			return array('result' => "Can not write to the directory. Note that it must be owned by the web server's user!");
		return array('result' => substr($line, strlen($msgbegin)));
	}

	updateRecordingList();
	$msgbegin = "Job number is ";
	if (strpos($line, $msgbegin) === 0) {
		return array(   'result' => null,
				'number' => intval(substr($line, strlen($msgbegin))));
	}
	return array('result'=>"Got unparsable output from $dvbsched:\n $output");
}

# Edits the job in the schedule. 
# First argument is the job to edit.
# Second argument is a job map, like the one returned in the getSchedule() list.
# 'number', 'user', 'dvbType' will be ignored. If 'file' does not start with
# a slash ('/'), $recordDir from config.php will be prepended. If 'endTime' is
# set, it will be taken, otherwise 'duration'.
# Returns an array map with two values:
#   'result': null on success, or an error message on failure.
#   'number': the new number of the job after editing. Note that it will almost
#             always change, even if the operation failed
function editJob($number, $job) {
	$oldJob = getJob($number);
	if (!$oldJob)
		return array('result'=>"Cannot get job $number for editing.",
			     'number'=>$number);

	$status = removeJob($number);
	if ($status)
		return array('result'=>"Error while removing old job: $status",
			     'number'=>$number);

	$status = addJob($job);
	if ($status['result']!=null) {
		$errmsg = "Error while adding new job: ".$status['result'];
		$status2 = addJob($oldJob);
		if ($status2['result']!=null) {
			$errmsg = $errmsg . "\nError while restoring old job: ".$status2['result']."\n The job is lost :(";
			return array('result'=>$errmsg,
				     'number'=>$number);
		}
		return array('result'=>$errmsg,
			     'number'=>$status2['number']);
	}
        return array('result'=>null,
                     'number'=>$status['number']);
}

# Returns a job array map with default values for creating a new job
# 'number' and 'user' are undefined.
function getDefaultJob() {
	global $dvbType, $convertToPsDefault, $PRIO_NORMAL;
	$channels = getChannels();
	$time = date("H:i");
	$duration = 75;
	return array(   'date' => date("Y-m-d"),
                        'time' => $time,
			'endTime' => addTime($time, $duration),
                        'dvbType' => $dvbType,
                        'channel' => $channels[0],
                        'convertToPs' => $convertToPsDefault,
                        'duration' => $duration,
                        'priority' => $PRIO_NORMAL,
                        'file' => date("_m_d").'.mpg',
                        'extra' => ''
                );
}

# Returns a list of the all recorded files in $recordDir, as 
# relative file name string array
function getAllRecordedFiles() {
	global $recordDir;
	$filesRaw = glob($recordDir."{*.mpg,*.ts}", GLOB_BRACE);
	$filenames = array();
	foreach ($filesRaw as $fileRaw) {
                $npos = strpos($fileRaw, $recordDir);
                if ($npos === 0)
                        $filenames[] = substr($fileRaw, strlen($recordDir));
		else
			$filenames[] = $fileRaw;
	}
	return $filenames;
}

# Returns the "base" name of a recorded file. The base name
# - has not extension (.ts or .mpeg)
# - the postfixes '_cut', '(cut)' and ' cut' will be removed
# - Spaces at beginning and end will be removed
function getBaseName($filename) {
	$fl = strlen($filename);
	if (substr($filename, -3, 3) == '.ts')
		$f = substr($filename, 0, $fl-3);
	else if (substr($filename, -4, 4) == '.mpg')
		$f = substr($filename, 0, $fl-4);
	else 
		$f = $filename;

	$fl = strlen($f);
        if (substr($f, -4, 4) == '_cut')
                $f = substr($f, 0, $fl-4);
        else if (substr($f, -5, 5) == '(cut)')
                $f = substr($f, 0, $fl-5);
        else if (substr($f, -4, 4) == ' cut')
                $f = substr($f, 0, $fl-4);

	return trim($f);
}

# Internal function to parse the recordingList
# Returns array map, as described in getRecordingList()
function parseRecordingListLine($line) {
	$r = array();

        $cs = explode(" ", $line, 8);
	return array(	'number' => $cs[0],
			'date' => $cs[1],
			'time' => $cs[2],
			'endTime' => $cs[3],
			'duration' => $cs[4],
			'channel' => unescapeWS($cs[5]),
			'file' => unescapeWS($cs[6]),
			'fileBase' => getBaseName(unescapeWS($cs[6])),
			'extra' => unescapeWS($cs[7]),
			'size' => 0,
			'fullInfo' => true);
}

# Creates / Updates the list of all (prior and future) recordings,
# which is stored in $recordDir/.webdvbd-recording-list
function updateRecordingList() {
	global $recordingListPath;

	# 1. Read list from disk (schedules without files are already filtered out)
	$recordingList = getRecordingList();

	# 2. Add entries for all scheduled jobs
	$schedule = getSchedule();
	foreach ($schedule as $job)
		$recordingList[] = array(	
			'number' => $job['number'],
			'date' => $job['date'],
			'time' => $job['time'],
			'endTime' => $job['endTime'],
			'duration' => $job['duration'],
			'channel' => $job['channel'],
			'file' => $job['file'],
			'fileBase' => getBaseName($job['file']),
			'extra' => $job['extra'],
			'size' => 0,
			'fullInfo' => true);

	# 3. Save on disk
	$output = "";
	foreach ($recordingList as $r)
		if ($r['fullInfo'])
			$output .= $r['number'] . ' ' . $r['date'] . ' ' .
        	    		   $r['time'] . ' ' . $r['endTime'] . ' ' .
				   $r['duration'] . ' ' . 
				   escapeWS($r['channel']) . ' ' .
				   escapeWS($r['file']) . ' ' .
				   escapeWS($r['extra']) . "\n";
	$f = fopen($recordingListPath, 'w');
	if ($f) { // no error handling :(
		fwrite($f, $output);
		fclose($f);
	}

	return $recordingList;
}


# Internal function to read the recording list from disk
# Returns an array like getRecordingList(), but returns entries
# as stored in the .webdvbd-recording-list file, unsynced to the
# content of the $recordDir
function getRecordingListRaw() {
	global $recordingListPath;
	if (is_readable($recordingListPath))
		$unparsedList = file($recordingListPath);
	else
		$unparsedList = array();
	$recordingList = array();
	foreach($unparsedList as $line)
		$recordingList[] = parseRecordingListLine($line);
	return $recordingList;
}

# Returns a list of all recordings that have already been saved on disk. 
# Each recording is represented by an array map:
#  'file' => Filename of the recording (without path)
#  'fileBase' => Base-Filename (see getBaseName())
#  'size' => File size in bytes
#  'fullInfo' => if true, the following fields are available:
#    'number' => Number of the job
#    'date' => Date of recording (same format as in getSchedule())
#    'time' => Start time of recording (same format as in getSchedule())
#    'endTime' => End time of recording (same format as in getSchedule())
#    'duration' => Duration of the recording in minutes
#    'channel' => The channel that has been recorded
#    'extra' => Extra-Description (same as in getSchedule())
function getRecordingList() {
	global $recordDir;

	$rawList = getRecordingListRaw();
	$rawListByBase = array();
	foreach ($rawList as $entry)
		$rawListByBase[$entry['fileBase']] = $entry;

	$fileList = getAllRecordedFiles();
	$list = array();
	foreach ($fileList as $file) {
		$base = getBaseName($file);
		if ($rawListByBase[$base]) 
			$recording = $rawListByBase[$base];
		else
			$recording = array(
					'fileBase' => $base,
					'fullInfo' => false);
		$recording['file'] = $file;
		if (is_file($recordDir . $file)) {
			$rfstat = stat($recordDir . $file);
			$recording['size'] = $rfstat[7];
		}
		$list[] = $recording;
	}
	return $list;
}

# Returns the recording map of a specific file from a recording list, 
# as returned by getRecordingList(). Returns null if not found.
function getRecording($recordingList, $filename) {
	foreach ($recordingList as $recording)
		if ($recording['file'] == $filename)
			return $recording;
	return null;
}

# Translates the priority (as number) into a string.
function translatePriority($priority) {
	global $PRIO_LOW, $PRIO_NORMAL, $PRIO_HIGH;
	if ($priority == $PRIO_LOW)
		return "low";
	else if ($priority == $PRIO_NORMAL)
		return "normal";
	else if ($priority == $PRIO_HIGH)
		return "high";
	return $priority;
}
?>
